"""
Support functions for converting py values to sqream compatible values
and vice versa
"""

from __future__ import annotations

import numpy as np
import pandas as pd
from datetime import datetime, date
from decimal import Decimal, getcontext
from math import floor, ceil, pow
from typing import Union
from .globals import SQREAM_OLDEST_DATE_LONG, SQREAM_OLDEST_DATETIME_LONG, SQREAM_OLDEST_DATETIME2, TYPE_MAPPER
from struct import unpack


def pad_dates(num):
    return ('0' if num < 10 else '') + str(num)


def sq_date_to_py_date(sqream_date, is_null=False, date_convert_func=date):

    if is_null:
        return None

    year = (10000 * sqream_date + 14780) // 3652425
    intermed_1 = 365 * year + year // 4 - year // 100 + year // 400
    intermed_2 = sqream_date - intermed_1
    if intermed_2 < 0:
        year = year - 1
        intermed_2 = sqream_date - (365 * year + year // 4 - year // 100 +
                                    year // 400)
    intermed_3 = (100 * intermed_2 + 52) // 3060

    year = year + (intermed_3 + 2) // 12
    month = int((intermed_3 + 2) % 12) + 1
    day = int(intermed_2 - (intermed_3 * 306 + 5) // 10 + 1)

    return date_convert_func(year, month, day)


def sq_datetime_to_py_datetime(sqream_datetime, is_null=False, dt_convert_func=datetime):
    ''' Getting the datetime items involves breaking the long into the date int and time it holds
        The date is extracted in the above, while the time is extracted here  '''

    if is_null:
        return None

    date_part = sqream_datetime >> 32
    time_part = sqream_datetime & 0xffffffff
    date_part = sq_date_to_py_date(date_part, is_null=is_null)

    if date_part is None:
        return None

    msec = time_part % 1000
    sec = (time_part // 1000) % 60
    mins = (time_part // 1000 // 60) % 60
    hour = time_part // 1000 // 60 // 60
    return dt_convert_func(date_part.year, date_part.month, date_part.day,
                           hour, mins, sec, msec * int(pow(10, 3)))  # native python datetime has 6 digits precision (microseconds)
    # while sqream's datetime works with 3 digits precision (milliseconds)


def sq_datetime2_to_numpy_datetime64(datetime2_as_bytes: bytes, is_null: bool = False) -> [Decimal, None]:
    """
    Converts SQream's datetime2 (ftTimestampTz which has nanosecond precision) to numpy's datetime64 (native python datetime supports microseconds, not nanoseconds).
    SQream's datetime2 stores 4 integers (each is 32 bits/4 bytes):
      - 1. date (year, month, day)
      - 2. time (with milliseconds precision)
      - 3. nanoseconds
      - 4. TimeZone

    The strategy is to unpack the 4 integers and construct numpy's datetime64 object which supports nanoseconds precision
    """

    if is_null:
        return None

    utc_offset_seconds, nano_seconds, time_as_int, date_as_int = unpack(TYPE_MAPPER.get_pack_code("ftTimestampTz"), datetime2_as_bytes)

    date_part = sq_date_to_py_date(date_as_int, is_null)
    if date_part is None:
        return None

    time_ns = sq_time_to_ns_since_midnight(time_as_int, nano_seconds)
    date_np = np.datetime64(date_part, 'D')

    datetime_ns = date_np.astype('datetime64[ns]') + np.timedelta64(time_ns, 'ns')
    offset_ns = np.timedelta64(utc_offset_seconds, 's').astype('timedelta64[ns]')

    return datetime_ns + offset_ns


def sq_time_to_ns_since_midnight(time_as_int, nano_seconds):
    """
    Convert SQream integer time and nanoseconds to total nanoseconds since midnight
    """

    hour = time_as_int // (1000 * 60 * 60)
    minutes = (time_as_int // (1000 * 60)) % 60
    seconds = (time_as_int // 1000) % 60
    milliseconds = time_as_int % 1000

    return (
            hour * 3600 * 10**9 +
            minutes * 60 * 10**9 +
            seconds * 10**9 +
            milliseconds * 10**6 +
            nano_seconds
    )


def numpy_datetime64_to_sq_datetime2(np_datetime: Union[np.datetime64, None]) -> tuple[int, int, int, int]:
    """
    Converts numpy's datetime64 to SQream's datetime2 representation as 4 integers that will later on will be packed in bytes.
    The output should contain 4 integers (each 4 bytes):
    - UTC offset in seconds
    - Nanoseconds (only the sub-millisecond part)
    - Time as int (including milliseconds)
    - Date as int
    """

    if np_datetime is None:
        return SQREAM_OLDEST_DATETIME2

    ts = pd.Timestamp(np_datetime)
    utc_offset_seconds = 0
    if ts.tzinfo is not None:
        utc_offset_seconds = int(ts.utcoffset().total_seconds())

    np_datetime = np_datetime.astype('datetime64[ns]')  # ensuring nano seconds precision

    date_part = np_datetime.astype('datetime64[D]')
    ns_since_midnight = (np_datetime - date_part).astype('timedelta64[ns]').astype(np.int64)

    total_milliseconds = ns_since_midnight // 1_000_000
    nano_seconds = int(ns_since_midnight % 1_000_000)  # keeps only sub-millisecond part (nanoseconds precision has 9 digits, we hold here the last 6)

    time_as_int = int(total_milliseconds)  # including milliseconds

    date_as_int = date_to_int(date_part.astype(datetime))

    return utc_offset_seconds, nano_seconds, time_as_int, date_as_int


def _get_date_int(year: int, month: int, day: int) -> int:
    """Convert year, month and day to integer compatible with SQREAM"""
    month: int = (month + 9) % 12
    year: int = year - month // 10
    return (
        365 * year + year // 4 - year // 100 + year // 400
        + (month * 306 + 5) // 10 + (day - 1)
    )


def date_to_int(dat: date) -> int:
    """Convert datetime.date to integer compatible with SQREAM interface"""
    # datetime is also supported because it is descendant of date
    # date_to_int(date(1900, 1, 1)) is 693901 which is the oldest date that
    # sqream supports, so for None use the same
    return SQREAM_OLDEST_DATE_LONG if dat is None else _get_date_int(*dat.timetuple()[:3])


def datetime_to_long(dat: datetime) -> int:
    """Convert datetime.datetime to integer (LONG) compatible with SQREAM"""
    if dat is None:
        # datetime_to_long(datetime(1900, 1, 1)) is 2980282101661696 which is
        # the oldest date that sqream supports, so for None use the same
        return SQREAM_OLDEST_DATETIME_LONG
    year, month, day, hour, minute, second = dat.timetuple()[:6]
    msec = dat.microsecond

    date_int: int = _get_date_int(year, month, day)
    time_int: int = 1000 * (hour * 3600 + minute * 60 + second) + round(msec / 1000)

    return (date_int << 32) + time_int


tenth = Decimal("0.1")
if getcontext().prec < 38:
    getcontext().prec = 38


def sq_numeric_to_decimal(bigint_as_bytes: bytes, scale: int, is_null=False) -> [Decimal, None]:
    if is_null:
        return None

    getcontext().prec = 38
    c = memoryview(bigint_as_bytes).cast('i')
    bigint = ((c[3] << 96) + ((c[2] & 0xffffffff) << 64) + ((c[1] & 0xffffffff) << 32) + (c[0] & 0xffffffff))

    return Decimal(bigint) * (tenth ** scale)


def decimal_to_sq_numeric(dec: Decimal, scale: int) -> int:  # returns bigint
    if getcontext().prec < 38:
        getcontext().prec = 38
    res = dec * (10 ** scale)
    return ceil(res) if res > 0 else floor(res)


def lengths_to_pairs(nvarc_lengths):
    ''' Accumulative sum generator, used for parsing nvarchar columns '''

    idx = new_idx = 0
    for length in nvarc_lengths:
        new_idx += length
        yield idx, new_idx
        idx = new_idx


def arr_lengths_to_pairs(text_lengths):
    """Generator for parsing ARRAY TEXT columns' data"""
    start = 0
    for length in text_lengths:
        yield start, length
        start = length + (8 - length % 8) % 8


def numpy_datetime_str_to_tup(numpy_dt):
    ''' '1970-01-01T00:00:00.699148800' '''

    numpy_dt = repr(numpy_dt).split("'")[1]
    date_part, time_part = numpy_dt.split('T')
    year, month, day = date_part.split('-')
    hms, ns = time_part.split('.')
    hour, mins, sec = hms.split(':')
    return year, month, day, hour, mins, sec, ns


def numpy_datetime_str_to_tup2(numpy_dt):
    ''' '1970-01-01T00:00:00.699148800' '''

    ts = (numpy_dt - np.datetime64('1970-01-01T00:00:00Z')) / np.timedelta64(1, 's')
    dt = datetime.utcfromtimestamp(ts)

    return dt.year, dt.month, dt.day



"""SQream Native Python API"""
import time
from datetime import datetime, date, time as t
from pysqream.connection import Connection
from pysqream.logger import log_and_raise, start_logging, stop_logging
from pysqream.server.connection_params import ConnectionParams


def enable_logs(log_path=None):
    start_logging(None if log_path is True else log_path)


def stop_logs():
    stop_logging()


def connect(host: str, port: int, database: str, username: str, password: str, clustered: bool = False,
            use_ssl: bool = False, service: str = "sqream", log: bool = False, **kwargs):
    """
    Connect to SQream database
    """

    conn_params = ConnectionParams(host, port, database, username, password, clustered, use_ssl, service)
    conn = Connection(conn_params, log=log, base_connection=True, allow_array=kwargs.get("allow_array", True))

    return conn


#  DB-API compatibility
#  -------------------
""" To fully comply to Python's DB-API 2.0 database standard. Ignore when using internally """


class _DBAPITypeObject:
    """DB-API type object which compares equal to all values passed to the constructor.
        https://www.python.org/dev/peps/pep-0249/#implementation-hints-for-module-authors
    """
    def __init__(self, *values):
        self.values = values

    def __eq__(self, other):
        return other in self.values


# Type objects and constructors required by the DB-API 2.0 standard
Binary = memoryview
Date = date
Time = t
Timestamp = datetime


STRING = "STRING"
BINARY = _DBAPITypeObject("BYTES", "RECORD", "STRUCT")
NUMBER = _DBAPITypeObject("INTEGER", "INT64", "FLOAT", "FLOAT64", "NUMERIC",
                          "BOOLEAN", "BOOL")
DATETIME = _DBAPITypeObject("TIMESTAMP", "DATE", "TIME", "DATETIME")
ROWID = "ROWID"


def DateFromTicks(ticks):
    return Date.fromtimestamp(ticks)


def TimeFromTicks(ticks):
    return Time(
        *time.localtime(ticks)[3:6]
    )  # localtime() returns a namedtuple, fields 3-5 are hr/min/sec


def TimestampFromTicks(ticks):
    return Timestamp.fromtimestamp(ticks)


# DB-API global parameters
apilevel = '2.0' 
threadsafety = 1  # Threads can share the module but not a connection
paramstyle = 'qmark'



import numpy as np
from mmap import mmap
from dataclasses import dataclass
from abc import ABC, abstractmethod
from typing import Any, List, Union
from pysqream.globals import ROWS_PER_FLUSH, TYPE_MAPPER


class PackingError(Exception):
    """
    Custom exception for packing errors
    """

    def __init__(self, message: str, col_idx: int = None, col_type: str = None):
        self.message = message
        if col_idx is not None and col_type is not None:
            self.message = f"Error packing column {col_idx + 1} of type '{col_type}', error is: {message}"
        super().__init__(self.message)


@dataclass
class ColumnMetadata:
    """
    Metadata for a column to be packed/unpacked
    """

    col_idx: int
    col_type: Union[str, list[str]]  # Support for both regular and array types. array is represented as list of two strings, example: ['ftArray', 'ftInt']
    size: int
    nullable: bool
    tvc: bool
    scale: int


@dataclass
class ColumnData:
    """
    Actual data received from server from fetch request.
    Used on unpack function
    """

    data_column: Union[memoryview, bytes]
    nullable: Union[memoryview, bool]
    array_lengths: Union[memoryview, None]
    true_nvarchar: Union[memoryview, bool]


class ColumnPacker(ABC):
    """
    Abstract base class for column packers
    """

    uses_standard_packing: bool = True

    def __init__(self, metadata: ColumnMetadata):
        self.metadata: ColumnMetadata = metadata
        self.buf_map: Union[mmap, None] = None
        self.buf_idx: int = 0

    @staticmethod
    def _is_null(nullable_byte: int) -> bool:
        return nullable_byte == 1

    def pack(self, data: Union[List[Any], np.ndarray]) -> bytes:
        """
        Template method that defines the packing algorithm structure.
        """

        if self.uses_standard_packing:
            self.allocate_buf_map()
            self.handle_nulls(data)
            return self._pack_data(data)

        return self._pack_custom(data)

    @abstractmethod
    def unpack(self, column_data: ColumnData) -> List[Any]:
        """
        Abstract method that subclasses must implement to define their specific unpacking extract logic.
        """

        pass

    @abstractmethod
    def _pack_data(self, data: List[Any]) -> bytes:
        """
        Abstract method that subclasses must implement to define their specific packing logic.
        """

        pass

    def _pack_custom(self, data: Union[List[np.ndarray], List[Union[List[Any], None]]]) -> bytes:
        """
        Custom packing logic for subclasses that don't use the standard flow.
        Only needs to be implemented if 'uses_standard_packing' is False.
        """

        raise NotImplementedError("Custom packing not implemented for this packer")

    def _cast_data_by_pack_code(self, column_data: ColumnData):
        column_data.data_column = column_data.data_column.cast(TYPE_MAPPER.get_pack_code(self.metadata.col_type))

    @staticmethod
    def _cast_data_to_bytes(column_data: ColumnData):
        column_data.data_column = column_data.data_column.tobytes()

    def allocate_buf_map(self):
        """
        Allocate bytes buffer which we send to the server
        """

        self.buf_map = mmap(-1, ((1 if self.metadata.nullable else 0) + (self.metadata.size if self.metadata.size != 0 else 104)) * ROWS_PER_FLUSH)

    def clear(self):
        """
        Clears the byte buffer
        """

        if self.buf_map:
            self.buf_map.close()

    def handle_nulls(self, data: List[Any]) -> int:
        """
        Handle null values in the column data.
        Writes nulls into the buffer and increments the buffer index
        """

        if not self.metadata.nullable:
            return 0

        idx = -1

        while True:
            try:
                idx = data.index(None, idx + 1)
            except ValueError:
                break
            self.buf_map.seek(self.buf_idx + idx)
            self.buf_map.write(b'\x01')
            data[idx] = TYPE_MAPPER.get_default_value(self.metadata.col_type, self.metadata.size)

        self.buf_idx = len(data)

    def write_packed_data(self, packed_data: Union[list, bytes]) -> bytes:
        """
        Helper method to write packed data to buffer
        """

        self.buf_map.seek(self.buf_idx)
        self.buf_map.write(bytearray(packed_data))
        self.buf_idx += len(packed_data)

        return self.buf_map[:self.buf_idx]



"""
Test for insertion array via network

This imply using syntax:
    # Get a new cursor
    cur = con.cursor()
    insert = "insert into perf values (?,?)"  # insert in 2 columns
    cur.executemany(insert, [
        (array_col1_row1, array_col2_row1),  # row1
        (array_col1_row2, array_col2_row2),  # row2
    ])
    # Close this cursor
    cur.close()
"""
from datetime import date, datetime
from decimal import Decimal

import pytest
import numpy as np  # numpy is currently in requirements of the package

from pysqream.utils import DataError
from pysqream.globals import ROWS_PER_FLUSH

from tests.utils import ALL_TYPES, SIMPLE_VALUES, ensure_empty_table, select

TEMP_TABLE = "test_array_network_insert_temp"


WRONG_TYPES_VALUES = [
    [True, False, "What?"],
    [1, 255, 255.1],
    [256, 32767, 3.5],
    [32768, 2147483647, 1.0123],
    [2147483648, 9223372036854775807, 1.5],

    [3.1410000324249268, 5.315000057220459, "a"],
    [0.000003, 101.000026, "b"],

    [Decimal('0.12324567890123456789012345678901234567'), "Don't do that"],
    [Decimal('131235.1232'), [Decimal('131235.1232')]],

    [date(1955, 11, 5), '9999-12-31'],  # Does not accept strings
    [datetime(1955, 11, 5, 1, 24), date(1955, 11, 5)],
    [np.datetime64('1955-11-05T01:24:00.000000000'), datetime(1955, 11, 5, 1, 24), date(1955, 11, 5)],
    ["Kiwis have tiny wings, but cannot fly.", 12331, b'No bytes!'],
]


@pytest.fixture(name='cursor')
def cursor_with_arrays_allowed(sqream_cursor):
    """Redefined cursor fixture that enables arrays for this tests module"""
    sqream_cursor.allow_array = True
    yield sqream_cursor


@pytest.mark.parametrize("data_type, data", [
    ("BOOL", [True, False, True, None, True]),
    ("TINYINT", [2, None, 255]),  # less than 256 (1 byte)
    ("SMALLINT", [None, 32767, 256, 7]),
    ("INT", [2147483647, 32768, None, 56]),
    ("BIGINT", [9223372036854775807, 2147483647, 1, None]),
    ("REAL", [3.1410000324249268, None, 5.315000057220459, 1.0]),
    ("DOUBLE", [0.000003, None, 10.000001, 101.000026, 2.3]),
    ("NUMERIC(38,38)", [
        Decimal('0.12324567890123456789012345678901234567'), None]),
    ("NUMERIC(12,4)", [Decimal('131235.1232'), None]),
    ("DATE", [date(1955, 11, 5), None, date(9999, 12, 31)]),
    # ("DATETIME", [datetime(1955, 11, 5, 1, 24),
    #               datetime(9999, 12, 31, 23, 59, 59, 999), None]),
    ("TEXT", [None, "Kiwis have tiny wings, but cannot fly.", "", "xXx"]),
    ("DATETIME2", [np.datetime64('1955-11-05T01:24:00.000000000'), np.datetime64('9999-12-31T23:59:59.999000000'), None])
])
def test_insert_most_types(cursor, data_type, data):
    """Test insert most types except DATETIME"""
    ensure_empty_table(cursor, TEMP_TABLE, f"d {data_type}[]")

    cursor.executemany(f"insert into {TEMP_TABLE} values (?)", [(data,)])

    assert select(cursor, TEMP_TABLE) == [(data, )]


@pytest.mark.parametrize("year, month, day, hour, minute, sec, mcrsec", [
    (1955, 11, 5, 1, 24, 0, 0),
    (9999, 12, 31, 23, 59, 59, 999),
])
def test_insert_datetime(cursor, year, month, day, hour, minute, sec, mcrsec):
    """
    Test insert array of datetime works such buggy as DATATIME itself

    Should move to test_insert_most_types after fixing bugs:
    Bug1 (SELECT) https://sqream.atlassian.net/browse/SQ-13967
    Bug2 (INSERT) https://sqream.atlassian.net/browse/SQ-13969
    """
    # Passing many argument allows to see datetime in tests explicitly, so
    # pylint: disable=too-many-arguments
    ensure_empty_table(cursor, TEMP_TABLE, "d DATETIME[]")

    data = datetime(year, month, day, hour, minute, sec, mcrsec)
    data2 = datetime(year - 1, month, day, hour, minute, sec, mcrsec)
    cursor.executemany(
        f"insert into {TEMP_TABLE} values (?)", [([data, data2], )])

    # adjust to working in current bugs environment
    res1 = datetime(year, month, day, hour, minute, sec, round(mcrsec / 1000) * 1000)
    res2 = datetime(year - 1, month, day, hour, minute, sec, round(mcrsec / 1000) * 1000)
    assert select(cursor, TEMP_TABLE) == [([res1, res2], )]


@pytest.mark.parametrize("_type", ALL_TYPES)
def test_insert_all_types_send_none(cursor, _type):
    """Test that None send appropriate values for nullable columns"""
    ensure_empty_table(cursor, TEMP_TABLE, f"d {_type}[] null")

    data = [(None,)] * 15
    cursor.executemany(f"insert into {TEMP_TABLE} values (?)", data)
    assert select(cursor, TEMP_TABLE) == data


@pytest.mark.parametrize("_type", ALL_TYPES)
def test_insert_all_types_send_none_for_not_nullable(cursor, _type):
    """Test that DataError raises in case of sending Nones to not nullable"""
    ensure_empty_table(cursor, TEMP_TABLE, f"d {_type}[] not null")

    with pytest.raises(DataError):
        cursor.executemany(f"insert into {TEMP_TABLE} values (?)", [(None,)])


@pytest.mark.parametrize("size", [2, 10, 50])
@pytest.mark.parametrize("_type", ALL_TYPES)
def test_insert_fixed_size_array_works(cursor, _type, size):
    """Test that insert works with lists of length less or equal to defined"""
    ensure_empty_table(cursor, TEMP_TABLE, f"d {_type}[{size}]")

    data = [([None] * i, ) for i in range(size + 1)] * 15
    cursor.executemany(f"insert into {TEMP_TABLE} values (?)", data)
    assert select(cursor, TEMP_TABLE) == data


@pytest.mark.skip("Network insertion skips it silently SQ-13979")
@pytest.mark.parametrize("size", [2, 10, 50])
@pytest.mark.parametrize("_type", ALL_TYPES)
def test_insert_fixed_size_array_exceed_range_raises(cursor, _type, size):
    """Test that insert raises with lists of length greater than defined"""
    ensure_empty_table(cursor, TEMP_TABLE, f"d {_type}[{size - 1}]")
    data = [([None] * i, ) for i in range(size + 1)] * 15
    with pytest.raises(Exception):
        # Should not validate response
        cursor.executemany(f"insert into {TEMP_TABLE} values (?)", data)


@pytest.mark.parametrize("_type", ALL_TYPES)
def test_insert_all_types_send_empty_list(cursor, _type):
    """Test insertion of empty lists"""
    ensure_empty_table(cursor, TEMP_TABLE, f"d {_type}[]")
    data = [([],)] * 10
    cursor.executemany(f"insert into {TEMP_TABLE} values (?)", data)
    assert select(cursor, TEMP_TABLE) == data


@pytest.mark.parametrize("_type, val", zip(ALL_TYPES, SIMPLE_VALUES))
def test_insert_more_rows_than_per_flush(cursor, _type, val):
    """Test insertion works with sending rows by parts"""
    ensure_empty_table(cursor, TEMP_TABLE, f"d {_type}[]")
    data = [(val,)] * (ROWS_PER_FLUSH + 2)
    cursor.executemany(f"insert into {TEMP_TABLE} values (?)", data)
    res = select(cursor, TEMP_TABLE)
    # Do some lightweight checks before full compare to make it easier
    # both to debug and to provide comparison by pytest
    assert len(res) == len(data)
    assert res[0] == data[0]
    mid = int(ROWS_PER_FLUSH / 2)
    assert res[mid] == data[mid]
    assert res[-1] == data[-1]
    assert res == data


@pytest.mark.parametrize("_type, val", zip(ALL_TYPES, WRONG_TYPES_VALUES))
def test_inappropriate_type_raises(cursor, _type, val):
    """Test wrong python types raises DataError"""
    ensure_empty_table(cursor, TEMP_TABLE, f"d {_type}[]")
    data = [(val, )]
    with pytest.raises(DataError):
        cursor.executemany(f"insert into {TEMP_TABLE} values (?)", data)
    assert select(cursor, TEMP_TABLE) == []


@pytest.mark.parametrize("_type, val", zip(ALL_TYPES, SIMPLE_VALUES))
def test_numpy_arrays(cursor, _type, val):
    """Test that numpy data types works as well as python built-ins"""
    ensure_empty_table(cursor, TEMP_TABLE, f"d {_type}[]")
    data = [(np.array(val),)]
    cursor.executemany(f"insert into {TEMP_TABLE} values (?)", data)
    res = select(cursor, TEMP_TABLE)
    assert res == [(val,)]



import pytest
import struct
import numpy as np
from pysqream import casting
from datetime import datetime
from tests.test_base import TestBaseWithoutBeforeAfter


class TestDatetimeUnitTest(TestBaseWithoutBeforeAfter):

    def test_zero_date(self):
        with pytest.raises(Exception, match="year 0 is out of range"):
            casting.sq_date_to_py_date(0, is_null=False)

    def test_zero_datetime(self):
        with pytest.raises(Exception, match="year 0 is out of range"):
            casting.sq_datetime_to_py_datetime(0, is_null=False)

    def test_negative_date(self):
        with pytest.raises(Exception, match="year -9 is out of range"):
            casting.sq_date_to_py_date(-3000, is_null=False)

    def test_negative_datetime(self):
        with pytest.raises(Exception, match="year 0 is out of range"):
            casting.sq_datetime_to_py_datetime(-3000, is_null=False)

    def test_null_date(self):
        res = casting.sq_date_to_py_date(-3000, is_null=True)
        assert res is None, f"Excepted to get None, but got [{res}]"

    def test_null_datetime(self):
        res = casting.sq_datetime_to_py_datetime(-3000, is_null=True)
        assert res is None, f"Excepted to get None, but got [{res}]"

    insert_data = [
        datetime(2015, 12, 24, 13, 11, 23, 0),
        datetime(2015, 12, 31, 23, 59, 59, 999),
        datetime(2015, 12, 24, 13, 11, 23, 1000),
        datetime(2015, 12, 24, 13, 11, 23, 1500),
        datetime(2015, 12, 31, 23, 59, 59, 2000),
    ]
    expected_result = [
        datetime(2015, 12, 24, 13, 11, 23),
        datetime(2015, 12, 31, 23, 59, 59, 1000),
        datetime(2015, 12, 24, 13, 11, 23, 1000),
        datetime(2015, 12, 24, 13, 11, 23, 2000),
        datetime(2015, 12, 31, 23, 59, 59, 2000),
    ]

    @pytest.mark.parametrize("input_datetime, expected_result", zip(insert_data, expected_result))
    def test_datetime_milliseconds_round(self, input_datetime, expected_result):
        """test case for SQ-13969"""

        dt_as_long = casting.datetime_to_long(input_datetime)
        long_as_dt = casting.sq_datetime_to_py_datetime(dt_as_long)

        assert long_as_dt == expected_result


class TestDatetime2UnitTest(TestBaseWithoutBeforeAfter):
    def test_sq_time_to_ns_since_midnight(self):
        # Test cases with different time components
        test_cases = [
            # time_as_int, nano_seconds, expected_ns
            (0, 0, 0),  # midnight, no nanos
            (1000, 456789, 1_000_456_789),  # 1 second + nanos
            (86399123, 456789, 86399123_000_000 + 456789),  # 23:59:59.123 + nanos
        ]

        for time_as_int, nano_seconds, expected_ns in test_cases:
            result = casting.sq_time_to_ns_since_midnight(time_as_int, nano_seconds)
            assert result == expected_ns

    def test_numpy_datetime64_to_sq_datetime2(self):
        # Test regular datetime conversion
        dt = np.datetime64('1997-12-31T23:59:59.123456789', 'ns')
        utc_offset, nanos, time_int, date_int = casting.numpy_datetime64_to_sq_datetime2(dt)

        assert utc_offset == 0
        assert nanos == 456789  # Only sub-millisecond part
        assert time_int == 86399123  # 23:59:59.123
        assert date_int == 729694  # 1997-12-31

        # Test None input
        result = casting.numpy_datetime64_to_sq_datetime2(None)
        assert result == casting.SQREAM_OLDEST_DATETIME2

        # Test midnight
        dt = np.datetime64('2000-01-01T00:00:00', 'ns')
        utc_offset, nanos, time_int, date_int = casting.numpy_datetime64_to_sq_datetime2(dt)
        assert time_int == 0
        assert nanos == 0

    def test_sq_datetime2_to_numpy_datetime64(self):
        # Test case using example bytes
        example_bytes = b'\x00\x00\x00\x00U\xf8\x06\x00\x93X&\x05^"\x0b\x00'
        result = casting.sq_datetime2_to_numpy_datetime64(example_bytes, False)

        # Convert back to verify
        components = casting.numpy_datetime64_to_sq_datetime2(result)
        datetime2_bytes = struct.pack('<iiii', *components)
        assert datetime2_bytes == example_bytes

        # Test null case
        assert casting.sq_datetime2_to_numpy_datetime64(example_bytes, True) is None

    def test_roundtrip_conversion(self):
        """Test that converting to SQream format and back gives the same datetime"""
        original_dt = np.datetime64('1997-12-31T23:59:59.123456789', 'ns')

        # Convert to SQream format
        components = casting.numpy_datetime64_to_sq_datetime2(original_dt)
        datetime2_bytes = struct.pack('<iiii', *components)

        # Convert back
        result_dt = casting.sq_datetime2_to_numpy_datetime64(datetime2_bytes, False)

        assert str(original_dt) == str(result_dt)

    def test_datetime2_timezone_conversions(self):
        """
        checking UTC time, Positive time zone offsets and Negative time zone offsets
        Verifies timezone information is preserved through conversion
        """
        test_cases = [
            np.datetime64('1970-01-17T18:56:04.205739Z', 'ns'),

            np.datetime64('1970-01-17T18:56:04.205739+02:00', 'ns'),
            np.datetime64('1970-01-17T18:56:04.205739+05:30', 'ns'),

            np.datetime64('1970-01-17T18:56:04.205739-05:00', 'ns'),
            np.datetime64('1970-01-17T18:56:04.205739-08:00', 'ns')
        ]

        for dt in test_cases:
            components = casting.numpy_datetime64_to_sq_datetime2(dt)
            datetime2_bytes = struct.pack('<iiii', *components)
            result_dt = casting.sq_datetime2_to_numpy_datetime64(datetime2_bytes, False)

            assert str(dt) == str(result_dt)
